class Constants:
    IMMEDIATE_MAX_WAIT_TIME = pow(2, 16) - 4
    """Max size of wait instruction immediates in Q1ASM programs. Max value allowed by
    assembler is 2**16-1, but this is the largest that is a multiple of 4 ns."""
    REGISTER_SIZE = pow(2, 32) - 1
    """Size of registers in Q1ASM programs."""
    NCO_PHASE_STEPS_PER_DEG = 1e9 / 360
    """The number of steps per degree for NCO phase instructions arguments."""
    NCO_FREQ_STEPS_PER_HZ = 4.0
    """The number of steps per Hz for the NCO set_freq instruction."""
    NCO_FREQ_LIMIT_STEPS = 2e9
    """The maximum and minimum frequency expressed in steps for the NCO set_freq instruction.
    For the minimum we multiply by -1."""
    NUMBER_OF_SEQUENCERS_QCM = 6
    """Number of sequencers supported by a QCM/QCM-RF in the latest firmware."""
    NUMBER_OF_SEQUENCERS_QRM = 6
    """Number of sequencers supported by a QRM/QRM-RF in the latest firmware."""
    NUMBER_OF_REGISTERS: int = 64
    """Number of registers available in the Qblox sequencers."""
    MAX_SAMPLE_SIZE_SCOPE_ACQUISITIONS: int = 16384
    """Maximal amount of scope trace acquisition datapoints returned."""
    MAX_SAMPLE_SIZE_WAVEFORMS: int = 16384
    """Maximal amount of samples in the waveforms to be uploaded to a sequencer."""
    GRID_TIME = 4  # ns
    """
    Clock period of the sequencers. All time intervals used must be multiples of this value.
    """
    LOOP_UNROLL_THRESHOLD = 4
    """
    Size above which loops have tolerable overhead
    """
    MIN_QCM_OFFSET_V = -2.5
    """
    Minimum offset for QCM
    """
    MAX_QCM_OFFSET_V = 2.5
    """
    Maximum offset for QCM
    """
    MIN_QCM_RF_OFFSET_MV = -84
    """
    Minimum offset for QCM-RF
    """
    MAX_QCM_RF_OFFSET_MV = 73
    """
    Maximum offset for QCM-RF
    """
    MIN_QRM_OFFSET_V = -0.09
    """
    Minimum offset for QRM
    """
    MAX_QRM_OFFSET_V = 0.09
    """
    Maximum offset for QRM
    """
    MIN_QRM_RF_OFFSET_V = -0.09
    """
    Minimum offset for QRM-RF
    """
    MAX_QRM_RF_OFFSET_V = 0.09
    """
    Maximum offset for QRM-RF
    """



import json
import os
from dataclasses import asdict
from datetime import datetime
from functools import reduce
from typing import Dict, List

import numpy as np
from qblox_instruments import Cluster, DummyScopeAcquisitionData
from qblox_instruments.qcodes_drivers.module import Module
from qblox_instruments.qcodes_drivers.sequencer import Sequencer

from qat.purr.backends.live_devices import ControlHardware, Instrument, LivePhysicalBaseband
from qat.purr.backends.qblox.codegen import QbloxPackage
from qat.purr.backends.qblox.config import (
    QbloxConfig,
    QcmConfigHelper,
    QcmRfConfigHelper,
    QrmConfigHelper,
    QrmRfConfigHelper,
)
from qat.purr.compiler.devices import (
    ChannelType,
    PhysicalChannel,
    PulseChannel,
    Qubit,
    Resonator,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QbloxResonator(Resonator):
    """
    A hack around QBlox acquisition to use the same pulse channel
    for MeasurePulse and Acquire instructions.
    """

    def get_measure_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.macq)

    def get_acquire_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.macq)


class QbloxPhysicalBaseband(LivePhysicalBaseband):
    def __init__(
        self,
        id_,
        frequency,
        if_frequency,
        instrument: Instrument,
        slot_idx: int,
        config: QbloxConfig,
    ):
        super().__init__(id_, frequency, if_frequency, instrument)
        self.slot_idx = slot_idx
        self.config = config  # TBD, is it worth flattening ?


class QbloxPhysicalChannel(PhysicalChannel):
    baseband: QbloxPhysicalBaseband

    @property
    def slot_idx(self):
        return self.baseband.slot_idx

    @property
    def config(self):
        return self.baseband.config

    def build_resonator(self, resonator_id, *args, **kwargs):
        """Helper method to build a resonator with default channels."""
        kwargs.pop("fixed_if", None)
        reson = QbloxResonator(resonator_id, self)
        reson.create_pulse_channel(ChannelType.macq, *args, fixed_if=True, **kwargs)
        reson.default_pulse_channel_type = ChannelType.macq
        return reson

    def build_qubit(
        self,
        index,
        resonator,
        drive_freq,
        second_state_freq=None,
        channel_scale=(1.0e-8 + 0.0j),
        measure_amp: float = 1.0,
        fixed_drive_if=False,
        qubit_id=None,
    ):
        """
        Helper method tp build a qubit with assumed default values on the channels. Modelled after the live hardware.
        """
        qubit = Qubit(index, resonator, self, drive_amp=measure_amp, id_=qubit_id)
        qubit.create_pulse_channel(
            ChannelType.drive,
            frequency=drive_freq,
            scale=channel_scale,
            fixed_if=fixed_drive_if,
        )

        qubit.create_pulse_channel(
            ChannelType.second_state, frequency=second_state_freq, scale=channel_scale
        )

        qubit.measure_acquire["delay"] = 151e-9  # TOF

        return qubit


class QbloxControlHardware(ControlHardware):
    _driver: Cluster

    def __init__(
        self,
        dev_id: str = None,
        name: str = None,
        address: str = None,
        cfg: Dict = None,
    ):
        super().__init__(id_=dev_id or os.environ.get("QBLOX_DEV_ID"))
        self.name = name or os.environ.get("QBLOX_DEV_NAME")
        self.address = address or os.environ.get("QBLOX_DEV_IP")
        self.cfg = cfg
        self.dump_sequence = False
        self._resources: Dict[Module, Dict[PulseChannel, Sequencer]] = {}

    def allocate_resources(self, target: PulseChannel):
        module_id = target.physical_channel.slot_idx - 1  # slot_idx is in range [1..20]
        module: Module = self._driver.modules[module_id]
        allocations = self._resources.setdefault(module, {})
        if target in allocations:
            return module, allocations[target]

        total = set(target.physical_channel.config.sequencers.keys())
        allocated = set([seq.seq_idx for seq in allocations.values()])

        available = list(total - allocated)
        if available:
            sequencer: Sequencer = module.sequencers[available[0]]
            allocations[target] = sequencer
            return module, sequencer

        raise ValueError(f"No more available sequencers on module {module}")

    def _get_acquisitions(self, module: Module, sequencer: Sequencer):
        module.get_sequencer_status(sequencer.seq_idx, timeout=1)
        module.get_acquisition_status(sequencer.seq_idx, timeout=1)

        acquisitions = module.get_acquisitions(sequencer.seq_idx)
        for acq_name in acquisitions:
            module.store_scope_acquisition(sequencer.seq_idx, acq_name)

        return module.get_acquisitions(sequencer.seq_idx)

    def _prepare_config(self, package: QbloxPackage, sequencer: Sequencer):
        if package.target.fixed_if:  # NCO freq constant
            nco_freq = package.target.baseband_if_frequency
            lo_freq = package.target.frequency - nco_freq
        else:  # LO freq constant
            lo_freq = package.target.baseband_frequency
            nco_freq = package.target.frequency - lo_freq

        qblox_config = package.target.physical_channel.config

        module_config = qblox_config.module
        if module_config.lo.out0_en:
            module_config.lo.out0_freq = lo_freq
        if module_config.lo.out1_en:
            module_config.lo.out1_freq = lo_freq
        if module_config.lo.out0_in0_en:
            module_config.lo.out0_in0_freq = lo_freq

        # Sequencer config from the HwM
        hwm_seq_config = qblox_config.sequencers[sequencer.seq_idx]
        # Sequencer config from QAT IR
        pkg_seq_config = package.sequencer_config
        hwm_seq_config.nco.freq = nco_freq
        hwm_seq_config.square_weight_acq.integration_length = (
            pkg_seq_config.square_weight_acq.integration_length
        )

        return qblox_config

    def connect(self):
        if self._driver is None or not Cluster.is_valid(self._driver):
            self._driver: Cluster = Cluster(
                name=self.name,
                identifier=self.address,
                dummy_cfg=(
                    self.cfg if self.address is None else None
                ),  # Ignore dummy config if an address is given
            )
            self._driver.reset()
        log.info(self._driver.get_system_status())
        self.is_connected = True

    def disconnect(self):
        if self._driver is not None:
            try:
                self._driver.close()
                self.is_connected = False
            except BaseException as e:
                log.warning(
                    f"Failed to close instrument ID: {self.id} at: {self.address}\n{str(e)}"
                )

    def install(self, package: QbloxPackage):
        module, sequencer = self.allocate_resources(package.target)
        log.debug(f"Configuring module {module}, sequencer {sequencer}")
        config = self._prepare_config(package, sequencer)
        if module.is_qcm_type:
            if module.is_rf_type:
                QcmRfConfigHelper(config).configure(module, sequencer)
            else:
                QcmConfigHelper(config).configure(module, sequencer)
        elif module.is_qrm_type:
            if module.is_rf_type:
                QrmRfConfigHelper(config).configure(module, sequencer)
            else:
                QrmConfigHelper(config).configure(module, sequencer)

        sequence = asdict(package.sequence)
        if self.dump_sequence:
            filename = f"schedules/sequence_{module.slot_idx}_{sequencer.seq_idx}_@_{datetime.utcnow()}.json"
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, "w") as f:
                f.write(json.dumps(sequence))

        log.debug(f"Uploading sequence to {module}, sequencer {sequencer}")
        sequencer.sequence(sequence)

        return module, sequencer

    def set_data(self, qblox_packages: List[QbloxPackage]):
        self._resources.clear()
        for package in qblox_packages:
            self.install(package)

    def start_playback(self, repetitions: int, repetition_time: float):
        if not any(self._resources):
            raise ValueError("No resources allocated. Install packages first")

        results = {}
        for module, allocations in self._resources.items():
            for target, sequencer in allocations.items():
                if module.is_qrm_type:
                    sequencer.delete_acquisition_data(all=True)

        for module, allocations in self._resources.items():
            for target, sequencer in allocations.items():
                module.arm_sequencer(sequencer.seq_idx)

        for module, allocations in self._resources.items():
            for target, sequencer in allocations.items():
                module.start_sequencer(sequencer.seq_idx)

        for module, allocations in self._resources.items():
            for target, sequencer in allocations.items():
                if module.is_qrm_type:
                    result_id = target.physical_channel.id
                    if result_id in results:
                        raise ValueError(
                            "Two or more pulse channels on the same physical channel"
                        )
                    acquisitions = self._get_acquisitions(module, sequencer)
                    for acq_name, acq in acquisitions.items():
                        i = np.array(acq["acquisition"]["bins"]["integration"]["path0"])
                        q = np.array(acq["acquisition"]["bins"]["integration"]["path1"])
                        results[result_id] = (
                            i + 1j * q
                        ) / sequencer.integration_length_acq()

        self._resources.clear()
        return results

    def __getstate__(self) -> Dict:
        results = super(QbloxControlHardware, self).__getstate__()
        results["_driver"] = None
        results["_resources"] = {}
        return results

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._driver = None
        self._resources = {}
        self.is_connected = False


class DummyQbloxControlHardware(QbloxControlHardware):
    def _setup_dummy_acq_data(self, module, sequencer: Sequencer):
        if module.is_qrm_type:
            waveforms = sequencer.get_waveforms()
            dummy_data = reduce(lambda a, b: a + b, [a["data"] for a in waveforms.values()])
            dummy_data = [(z, z) for z in dummy_data / np.linalg.norm(dummy_data)]
            dummy_scope_acquisition_data = DummyScopeAcquisitionData(
                data=dummy_data, out_of_range=(False, False), avg_cnt=(0, 0)
            )
            module.set_dummy_scope_acquisition_data(
                sequencer=sequencer.seq_idx, data=dummy_scope_acquisition_data
            )

    def set_data(self, qblox_packages: List[QbloxPackage]):
        self._resources.clear()
        for package in qblox_packages:
            module, sequencer = self.install(package)

            self._setup_dummy_acq_data(module, sequencer)

    def start_playback(self, repetitions: int, repetition_time: float):
        if not any(self._resources):
            raise ValueError("No resources allocated. Install packages first")

        results = {}
        for module, allocations in self._resources.items():
            for target, sequencer in allocations.items():
                module.arm_sequencer(sequencer.seq_idx)
                module.start_sequencer(sequencer.seq_idx)

                if module.is_qrm_type:
                    result_id = target.physical_channel.id
                    if result_id in results:
                        raise ValueError(
                            "Two or more pulse channels on the same physical channel"
                        )
                    acquisitions = self._get_acquisitions(module, sequencer)
                    for acq_name, acq in acquisitions.items():
                        i = np.array(acq["acquisition"]["bins"]["integration"]["path0"])
                        q = np.array(acq["acquisition"]["bins"]["integration"]["path1"])
                        results[result_id] = (
                            i + 1j * q
                        ) / sequencer.integration_length_acq()

        self._resources.clear()
        return results



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd

import qiskit.ignis.verification.randomized_benchmarking as rb

from qat.purr.compiler.runtime import get_builder
from qat.purr.integrations.qasm import Qasm2Parser
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


def randomized_benchmarking(hardware, nseeds, *args, **kwargs):
    """
    Generate randomized benchmarks. Please see Qiskits ``randomized_benchmarking_seq``
    for argument documentation. Returns a dictionary holding the QuantumProcess that
    relates to the various benchmarking runs and a list of the sequence lengths.
    """
    # Due to the variable return value we can't directly unpack.
    qiskit_results = rb.randomized_benchmarking_seq(nseeds, *args, **kwargs)
    if len(qiskit_results) > 2:
        raise NotImplementedError("Only standard rb currently supported.")

    rb_circs = qiskit_results[0]
    seq_lengths = qiskit_results[1]
    results = dict()
    index = 0
    for seed_iter in rb_circs:
        circuit_list = []
        for circuit in seed_iter:
            qasm = circuit.qasm()
            circuit_list.append(Qasm2Parser().parse(get_builder(hardware), qasm))
        results[index] = circuit_list
        index = index + 1

    return results, seq_lengths



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from __future__ import annotations

import inspect
import re
import sys
from enum import Enum, Flag, IntEnum, auto
from typing import List, Optional

from qat.purr.utils.serializer import json_dumps, json_loads


class InlineResultsProcessing(Flag):
    """
    Results transforms applied directly to the read-out value on the QPU. In most
    situations applied post-execution, but can also be interwoven.
    """

    # Raw readout from the QPU. Normally Z-axis values for each shot.
    Raw = auto()

    # Shot results averaged out to get a single 0/1 value for each qubit.
    Binary = auto()

    # Return the values in numpy arrays not Python lists.
    NumpyArrays = auto()

    Experiment = Raw | NumpyArrays
    Program = Binary

    def __repr__(self):
        return self.name


class ResultsFormatting(Flag):
    # Transforms each shot into binary then counts the instances.
    # Example for two qubits: { '00': 15, '01': 2524, '10': 250, '11': 730 }
    BinaryCount = auto()

    # Change results value based on conditions for ease-of-use. Set as a flag because
    # it means that return values from execution may change format unexpectedly, so
    # should have a way to disable it for certain uses.
    DynamicStructureReturn = auto()

    # If your qubit results are lists of binary, squash to one string representation.
    # Changes 1: [1, 0, 0, 1] to 1: '1001'.
    # Only works when Binary format is returned.
    SquashBinaryResultArrays = auto()

    def __repr__(self):
        return self.name


class QuantumResultsFormat:
    def __init__(self):
        self.format: Optional[InlineResultsProcessing] = None
        self.transforms: Optional[ResultsFormatting] = (
            ResultsFormatting.DynamicStructureReturn
        )

    def raw(self) -> QuantumResultsFormat:
        self.format = InlineResultsProcessing.Raw
        return self

    def binary(self) -> QuantumResultsFormat:
        self.format = InlineResultsProcessing.Binary
        return self

    def binary_count(self):
        """
        Returns a count of each instance of measured qubit registers.
        Switches result format to raw.
        """
        self.transforms = (
            ResultsFormatting.BinaryCount | ResultsFormatting.DynamicStructureReturn
        )
        self.format = InlineResultsProcessing.Raw
        return self

    def squash_binary_result_arrays(self):
        """
        Squashes binary result list into a singular bit string. Switches results to
        binary.
        """
        self.transforms = (
            ResultsFormatting.SquashBinaryResultArrays
            | ResultsFormatting.DynamicStructureReturn
        )
        self.format = InlineResultsProcessing.Binary
        return self

    def __contains__(self, other):
        if isinstance(other, ResultsFormatting):
            return self.transforms.__contains__(other)
        elif isinstance(other, InlineResultsProcessing):
            return self.format.__contains__(other)
        return False

    def __or__(self, other):
        if isinstance(other, ResultsFormatting):
            self.transforms = self.transforms.__or__(other)
        elif isinstance(other, InlineResultsProcessing):
            self.format = self.format.__or__(other)
        return self

    def __and__(self, other):
        if isinstance(other, ResultsFormatting):
            self.transforms = self.transforms.__and__(other)
        elif isinstance(other, InlineResultsProcessing):
            self.format = self.format.__and__(other)
        return self

    def __xor__(self, other):
        if isinstance(other, ResultsFormatting):
            self.transforms = self.transforms.__xor__(other)
        elif isinstance(other, InlineResultsProcessing):
            self.format = self.format.__xor__(other)
        return self

    def __repr__(self):
        return f"Format: {str(self.format)}. Transforms: {str(self.transforms)}."

    def __eq__(self, other):
        return self.format == other.format and self.transforms == other.transforms


class TketOptimizations(Flag):
    """Flags for the various Tket optimizations we can apply."""

    Empty = auto()
    DefaultMappingPass = auto()
    FullPeepholeOptimise = auto()
    ContextSimp = auto()
    DirectionalCXGates = auto()
    CliffordSimp = auto()
    DecomposeArbitrarilyControlledGates = auto()
    # EulerAngleReduction = auto()
    GlobalisePhasedX = auto()
    # GuidedPauliSimp = auto()
    KAKDecomposition = auto()
    # OptimisePhaseGadgets = auto()
    # PauliSimp = auto()
    # PauliSquash = auto()
    PeepholeOptimise2Q = auto()
    RemoveDiscarded = auto()
    RemoveBarriers = auto()
    RemoveRedundancies = auto()
    ThreeQubitSquash = auto()
    SimplifyMeasured = auto()

    One = DefaultMappingPass | DirectionalCXGates
    Two = One | FullPeepholeOptimise | ContextSimp


class QiskitOptimizations(Flag):
    """Flags for the various Qiskit optimizations we can apply."""

    Empty = auto()


class QatOptimizations(Flag):
    """Flags for the various Qat optimizations we can apply."""

    Empty = auto()


class MetricsType(Flag):
    Empty = auto()

    # Returns circuit after optimizations have been run.
    OptimizedCircuit = auto()

    # Count of transformed instructions after all forms of optimizations have
    # been performed.
    OptimizedInstructionCount = auto()

    # Set of basic metrics that should be returned at all times.
    Default = OptimizedCircuit | OptimizedInstructionCount

    def is_composite(self):
        """
        Any flags that are only composed of other ones should be signaled here. This
        is used for automatic metric generation and whether to build/validate this
        particular value.
        """
        return self == self.Default or self == self.Empty

    def snake_case_name(self):
        """
        Generate the Python field name that'll be used to hold the results of this
        metric.
        """
        name = self.name
        name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


class ErrorMitigationConfig(Flag):
    Empty = auto()
    MatrixMitigation = auto()
    LinearMitigation = auto()


class ExperimentalFeatures:
    error_mitigation = ErrorMitigationConfig.Empty


class CompilerConfig:
    """
    Full settings for the compiler. All values are defaulted on initialization.

    If no explicit optimizations are passed then the default set of optimization for the
    language you're attempting to compile will be applied.
    """

    def __init__(
        self,
        repeats=None,
        repetition_period=None,
        results_format: QuantumResultsFormat = None,
        metrics=MetricsType.Default,
        active_calibrations=None,
        optimizations: "OptimizationConfig" = None,
        error_mitigation: ErrorMitigationConfig = None,
    ):
        self.repeats: Optional[int] = repeats
        self.repetition_period: Optional = repetition_period
        self.results_format: QuantumResultsFormat = results_format or QuantumResultsFormat()
        self.metrics: MetricsType = metrics
        self.active_calibrations: List[CalibrationArguments] = active_calibrations or []
        self.optimizations: Optional[OptimizationConfig] = optimizations
        self.error_mitigation: Optional[ErrorMitigationConfig] = error_mitigation

    def to_json(self):
        return json_dumps(self, serializable_types=get_serializable_types())

    def from_json(self, json: str):
        vars(self).update(
            vars(json_loads(json, serializable_types=get_serializable_types()))
        )
        return self

    @classmethod
    def create_from_json(cls, json: str):
        return CompilerConfig().from_json(json)

    def validate(self, hardware):
        from qat.purr.compiler.hardware_models import QuantumHardwareModel

        if (
            self.error_mitigation is not None
            and self.error_mitigation != ErrorMitigationConfig.Empty
        ):
            if isinstance(hardware, QuantumHardwareModel) and (
                hardware.error_mitigation is None
                or not hardware.error_mitigation.readout_mitigation
            ):
                raise ValueError("Error mitigation not calibrated on this device.")
            if ResultsFormatting.BinaryCount not in self.results_format:
                raise ValueError(
                    "Binary Count format required for readout error mitigation"
                )


class CalibrationArguments:
    """Base class for individual calibration arguments."""

    def to_json(self):
        return json_dumps(self)

    def from_json(self, json: str):
        self.from_dict(json_loads(json))

    def _get_field_names(self):
        """Get existing field names for all attributes and properties"""
        existing_field_names = set(vars(self).keys())
        # Include class properties
        class_type = type(self)
        for prop in dir(class_type):
            if isinstance(getattr(class_type, prop), property):
                existing_field_names.add(prop)

        return existing_field_names

    def from_dict(self, dict_values):
        """
        Loads this dictionary into the arguments. Throws if key dosen't exist on the
        object.
        """
        valid_names = self._get_field_names()
        invalid_fields = [val for val in dict_values.keys() if val not in valid_names]
        if any(invalid_fields):
            raise ValueError(
                f"Field(s) {','.join(invalid_fields)} are not valid for "
                f"{self.__class__.__name__}."
            )

        vars(self).update(dict_values)


class Languages(IntEnum):
    Empty, Qasm2, Qasm3, QIR = range(4)

    def __repr__(self):
        return self.name


class OptimizationConfig:
    """
    Base class for instantiated optimizations as well as mix-in classes. Built this way
    so we can mix and match optimization objects across multiple setups and languages
    without duplication.
    """

    def __init__(self):
        super().__init__()

    def default(self):
        """Apply default set of optimizations to the current set."""
        return self

    def disable(self):
        """Disable all optimizations."""
        return self

    def minimum(self):
        """
        Apply minimum working set for current optimizations.
        """
        return self

    def __contains__(self, item):
        return False


class Tket(OptimizationConfig):
    def __init__(self, tket_optimization=None):
        super().__init__()
        self.tket_optimizations: TketOptimizations = TketOptimizations.Empty
        if tket_optimization is not None:
            self.tket_optimizations = tket_optimization

    def default(self):
        self.tket_optimizations = TketOptimizations.One
        return self

    def disable(self):
        self.tket_optimizations = TketOptimizations.Empty
        return self

    def minimum(self):
        self.tket_optimizations = TketOptimizations.DefaultMappingPass
        return self

    def __contains__(self, item):
        if isinstance(item, TketOptimizations) and item in self.tket_optimizations:
            return True
        return super().__contains__(item)


class Qiskit(OptimizationConfig):
    def __init__(self):
        super().__init__()
        self.qiskit_optimizations: QiskitOptimizations = QiskitOptimizations.Empty

    def default(self):
        self.qiskit_optimizations = QiskitOptimizations.Empty
        return self

    def __contains__(self, item):
        if isinstance(item, QiskitOptimizations) and item in self.qiskit_optimizations:
            return True
        return super().__contains__(item)


class Qasm2Optimizations(Tket, Qiskit):
    def __init__(self):
        super().__init__()
        self.default()

    def __repr__(self):
        return f"Qiskit: {self.qiskit_optimizations}. Tket: {self.tket_optimizations}."


class Qasm3Optimizations(OptimizationConfig):
    pass


class QIROptimizations(OptimizationConfig):
    pass


def get_optimizer_config(lang: Languages) -> Optional[OptimizationConfig]:
    """
    Returns the optimization config for this particular language. None if no valid ones
    found.
    """
    if lang == Languages.Qasm2:
        return Qasm2Optimizations()
    elif lang == Languages.Qasm3:
        return Qasm3Optimizations()
    elif lang == Languages.QIR:
        return QIROptimizations()
    return None


def get_config(lang: Languages, **kwargs):
    """
    Helper method to build a compiler config for a particular language. Forwards
    keywords to the CompilerConfig constructor.
    """
    config = CompilerConfig(**kwargs)
    config.optimizations = get_optimizer_config(lang)
    return config


serializable_types_dict = None


def get_serializable_types():
    global serializable_types_dict

    if serializable_types_dict is not None:
        return serializable_types_dict

    serializable_types = {}

    def update_dict(type):
        if issubclass(type, Enum):
            serializable_types.update({f"<enum '{type.__module__}.{type.__name__}'>": type})
        else:
            serializable_types.update({str(type): type})

    def get_serializable_types_dict(type):
        update_dict(type)

        for subclass in type.__subclasses__():
            get_serializable_types_dict(subclass)

    types_in_module = inspect.getmembers(sys.modules[__name__], inspect.isclass)

    for name, typ in types_in_module:
        get_serializable_types_dict(typ)

    serializable_types_dict = serializable_types

    return serializable_types_dict



from typing import Dict

import numpy as np

from qat.purr.backends.live import LiveDeviceEngine, LiveHardwareModel
from qat.purr.backends.live_devices import ControlHardware
from qat.purr.backends.qblox.codegen import QbloxEmitter
from qat.purr.backends.utilities import get_axis_map
from qat.purr.compiler.emitter import QatFile
from qat.purr.compiler.execution import SweepIterator
from qat.purr.compiler.instructions import AcquireMode
from qat.purr.compiler.interrupt import NullInterrupt


class QbloxLiveHardwareModel(LiveHardwareModel):
    def __init__(self, control_hardware: ControlHardware = None):
        super().__init__(control_hardware)
        self._reverse_coercion()

    def create_engine(self):
        return QbloxLiveEngine(self)

    def _reverse_coercion(self):
        """
        HwM pickler coerces primitive types into strings. This is a temporary
        workaround to support in-memory, pickled, and un-pickled dict key types
        of the qblox config
        """
        for channel in self.physical_channels.values():
            config = channel.config
            config.sequencers = {int(k): v for k, v in config.sequencers.items()}


class QbloxLiveEngine(LiveDeviceEngine):
    def startup(self):
        if self.model.control_hardware is None:
            raise ValueError(f"Please add a control hardware first!")
        self.model.control_hardware.connect()

    def shutdown(self):
        if self.model.control_hardware is None:
            raise ValueError(f"Please add a control hardware first!")
        self.model.control_hardware.disconnect()

    def _execute_on_hardware(
        self, sweep_iterator: SweepIterator, package: QatFile, interrupt=NullInterrupt()
    ):
        if self.model.control_hardware is None:
            raise ValueError("Please add a control hardware first!")

        results = {}
        while not sweep_iterator.is_finished():
            sweep_iterator.do_sweep(package.instructions)

            position_map = self.create_duration_timeline(package.instructions)

            qblox_packages = QbloxEmitter().emit(package)
            aq_map = self.build_acquire_list(position_map)
            self.model.control_hardware.set_data(qblox_packages)

            repetitions = package.repeat.repeat_count
            repetition_time = package.repeat.repetition_period

            for aqs in aq_map.values():
                if len(aqs) > 1:
                    raise ValueError(
                        "Multiple acquisitions are not supported on the same channel in one sweep step"
                    )
                for aq in aqs:
                    physical_channel = aq.physical_channel
                    dt = physical_channel.sample_time
                    physical_channel.readout_start = aq.start * dt + aq.delay
                    physical_channel.readout_length = aq.samples * dt
                    physical_channel.acquire_mode_integrator = (
                        aq.mode == AcquireMode.INTEGRATOR
                    )

            playback_results: Dict[str, np.ndarray] = (
                self.model.control_hardware.start_playback(
                    repetitions=repetitions, repetition_time=repetition_time
                )
            )

            for channel, aqs in aq_map.items():
                if len(aqs) > 1:
                    raise ValueError(
                        "Multiple acquisitions are not supported on the same channel in one sweep step"
                    )
                for aq in aqs:
                    response = playback_results[aq.physical_channel.id]
                    response_axis = get_axis_map(aq.mode, response)
                    for pp in package.get_pp_for_variable(aq.output_variable):
                        response, response_axis = self.run_post_processing(
                            pp, response, response_axis
                        )
                    var_result = results.setdefault(
                        aq.output_variable,
                        np.empty(
                            sweep_iterator.get_results_shape(response.shape),
                            response.dtype,
                        ),
                    )
                    sweep_iterator.insert_result_at_sweep_position(var_result, response)

        return results



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
import time
import uuid
from typing import List

from qiskit import QuantumCircuit, assemble, transpile
from qiskit.providers.basicaer import BasicAerJob, QasmSimulatorPy
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData

from qat.purr.backends.realtime_chip_simulator import get_default_RTCS_hardware
from qat.purr.compiler.config import CompilerConfig
from qat.purr.utils.logger import get_default_logger
from qat.qat import execute_qasm

logger = get_default_logger()


def run_qasm_circuit(circuit):
    backend_sim = QasmSimulatorPy()
    transpiled_qc = transpile(circuit, backend_sim)
    result = backend_sim.run(transpiled_qc).result()

    def _results_to_binary(results_dictionary):
        """
        Key value from Qiskit is in hex, aka ``0x2: 242``.
        Transform to binary, sort by actual integer value.
        """
        binary = {bin(int(key, 16))[2:]: val for key, val in results_dictionary.items()}
        fill_value = max([len(val) for val in binary.keys()])
        return {
            key.zfill(fill_value): val
            for key, val in sorted(binary.items(), key=lambda kvp: int(kvp[0]))
        }

    # If we're a single value, just return, otherwise return list of results (or None).
    if not any(result.results):
        return None
    elif len(result.results) == 1:
        return _results_to_binary(result.results[0].data.counts)
    else:
        return [_results_to_binary(res.counts) for res in result.results]


def run_qasm_from_file(qasm_file: str):
    """Runs the passed-in QASM file against the QASM state simulator."""
    return run_qasm_circuit(QuantumCircuit.from_qasm_file(qasm_file))


def run_qasm(qasm_str: str):
    """Runs the passed-in QASM string against the QASM state simulator."""
    return run_qasm_circuit(QuantumCircuit.from_qasm_str(qasm_str))


class QatBackend(QasmSimulatorPy):
    """
    Basic qiskit backend built to run QASM on our own backends.

    TODO: Expand this to become a proper back-end, as I don't believe we need to inherit
        off the QASM simulator as it stands.
    """

    def __init__(self, hardware=None, comp_config=None, **fields):
        super().__init__(**fields)
        self.hardware = hardware
        self.comp_config = comp_config

    def run(self, qobj, **backend_options):
        if not isinstance(qobj, List):
            qobj = [qobj]

        if not all(isinstance(val, QuantumCircuit) for val in qobj):
            raise ValueError(
                "Experiment list contains non QuantumCircuit objects which we can't "
                "process at this time."
            )

        options_keys = set(dir(self.options))
        assembled_data = assemble(
            qobj,
            self,
            **{key: val for key, val in backend_options.items() if key in options_keys},
        )

        if self.comp_config is None:
            self.comp_config = CompilerConfig()
        self.comp_config.repeats = assembled_data.config.shots
        self.comp_config.results_format.binary_count()

        # Merge the Qiskit header/config info together with the experiment as
        # represented as a QASM string
        experiment_info = [
            (
                val.qasm(),
                assembled_data.experiments[i].config,
                assembled_data.experiments[i].header,
            )
            for i, val in enumerate(qobj)
        ]

        results = []
        full_start = time.time()
        status = "COMPLETED"
        try:
            for circ, exp_config, header in experiment_info:
                target_hardware = self.hardware or get_default_RTCS_hardware()
                start = time.time()
                results_data = execute_qasm(circ, target_hardware, self.comp_config)
                end = time.time()

                # TODO: Massage results data into a form that should be returned. As
                #   we're count it should be hex binary count representation, x01: 5024
                #   etc.

                # Below, solution for a single classical register. Not sure how qat
                # deals with multiple cregs.
                if len(header.creg_sizes) == 1:
                    creg_label, _ = header.creg_sizes[0]
                    counts_qat = results_data[creg_label]
                    counts_qiskit = dict(
                        (hex(int(key[::-1], 2)), value)
                        for (key, value) in counts_qat.items()
                    )
                    results.append(
                        ExperimentResult(
                            self.comp_config.repeats,
                            True,
                            ExperimentResultData(counts=counts_qiskit),
                            status="DONE",
                            header=header,
                            name=header.name,
                            time_taken=end - start,
                        )
                    )
                else:
                    results.append(
                        ExperimentResult(
                            self.comp_config.repeats,
                            True,
                            ExperimentResultData(counts=results_data),
                            status="DONE",
                            header=header,
                            name=header.name,
                            time_taken=end - start,
                        )
                    )
        except Exception as ex:
            # Code further up infinitely recalls this method until it gets a result, so
            # catch and propagate message.
            status = str(ex)

        full_end = time.time()
        job_id = str(uuid.uuid4())
        return BasicAerJob(
            self,
            job_id,
            Result(
                backend_name=self.name(),
                backend_version=self._configuration.backend_version,
                qobj_id=assembled_data.qobj_id,
                job_id=job_id,
                results=results,
                status=status,
                success=any(results) and all(val.success for val in results),
                time_taken=full_end - full_start,
                header=assembled_data.header,
            ),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
import sys
import time

from qat.purr.compiler.metrics import MetricsMixin
from qat.purr.utils.logger import LoggerLevel, get_default_logger

log = get_default_logger()


class _LogContextManager(MetricsMixin):
    def __init__(
        self,
        message: str,
        metric_type=None,
        metrics_collection=None,
        level=LoggerLevel.INFO,
    ):
        super().__init__()
        if "{" not in message:
            raise ValueError("Need one matching {} to inject duration time into.")

        self.message = message
        self.level: int = level.value
        self.metric_type = metric_type
        self.start_time = None
        self.compilation_metrics = metrics_collection

    def __enter__(self):
        self.start_time = time.time()

    def __exit__(self, *args, **kwargs):
        # Arbitrarily limiting to 6 decimal places as greater precision
        # shouldn't be necessary.
        total_time = round(time.time() - self.start_time, 6)

        if total_time > sys.float_info.epsilon:
            log.log(self.level, self.message.format(total_time), stacklevel=1)

        if self.metric_type is not None:
            self.record_metric(self.metric_type, total_time)


def log_duration(
    message, metric_type=None, metric_collection=None, log_level: LoggerLevel = None
):
    if metric_type is not None and metric_collection is None:
        raise ValueError("Can't have a metric to record but no collection.")

    return _LogContextManager(
        message, metric_type, metric_collection, level=log_level or LoggerLevel.INFO
    )
